<div class="panel-heading">Channels you have access in</div>
<div class="panel-body">
	{IF EXISTS CHANNEL_NAMES}
		<div>
		<h3 style="margin: -12px 0 20px 0;"><small>Choose a channel to access it's Settings, Access or Akick pages.</small></h3>
		{FOR CH,ECH IN CHANNEL_NAMES,ESCAPED_CHANNEL_NAMES}
			{IF EQ PAGE_NAME /chanserv/info}
			<a href="set?channel={ECH}" class="btn btn-sm btn-primary {IF EQ ECH ESCAPED_CHANNEL}disabled{END IF}" style="margin-bottom: 4px;">{CH}</a>
			{ELSE}
			<a href="{PAGE_NAME}?channel={ECH}" class="btn btn-sm btn-primary {IF EQ ECH ESCAPED_CHANNEL}disabled{END IF}" style="margin-bottom: 4px;">{CH}</a>
			{END IF}
		{END FOR}
		</div>
	{ELSE}
		<em>You don't have access in any channels.</em><br>
	{END IF}
</div>
