/*
 *
 * (C) 2005-2017 Anope Team
 * Contact us at team@anope.org
 *
 * Please read COPYING and README for further details.
 */

//{{NO_DEPENDENCIES}}
// Microsoft Developer Studio generated include file.
// Used by Win32GUI.rc
//
#define VER_ANOPE			1
#define MANIFEST_RESOURCE_ID		2
#define ICON_APP			129


// Next default values for new objects
//
#ifdef APSTUDIO_INVOKED
#ifndef APSTUDIO_READONLY_SYMBOLS
#define _APS_NO_MFC			1
#define _APS_NEXT_RESOURCE_VALUE	152
#define _APS_NEXT_COMMAND_VALUE		40061
#define _APS_NEXT_CONTROL_VALUE		1167
#define _APS_NEXT_SYMED_VALUE		104
#endif
#endif
