						<br>
					</div>
				</div>
			</div>
			<div class="row">
				<div class="col-lg-12">
					<div class="footer text-center">
						Anope IRC Services - &copy; 2013-2017 Anope Team - <a href="http://anope.org">http://anope.org</a>
					</div>
				</div>
			</div>
		</div>
	</body>
</html>
