{INCLUDE header.html}
{INCLUDE chanserv/chanlist.html}
		<div class="panel-heading">Channel Information</div>
		<div class="panel-body">
			{FOR M IN MESSAGES}
			<div class="alert alert-info">
				{M}<br>
			</div>
			{END FOR}

			{IF EXISTS OKAY}
				{IF EXISTS CAN_SET}
				<form method="post" action="/chanserv/set?channel={CHANNEL_ESCAPED}">
				{END IF}
				<table id="tableInfo" class="table table-hover">
					<tr>
						<td>Channel Name</td>
						<td>{CHANNEL}</td>
					</tr>
					{IF EXISTS FOUNDER}
					<tr>
						<td>Founder</td>
						<td>{FOUNDER}</td>
					</tr>
					{END IF}
					{IF EXISTS SUCCESSOR}
					<tr>
						<td>Succsesor</td>
						<td>{SUCCESSOR}</td>
					</tr>
					{END IF}
					<tr>
						<td>Time registered</td>
						<td>{TIME_REGISTERED}</td>
					</tr>
					<tr>
						<td>Last used</td>
						<td>{LAST_USED}</td>
					</tr>
					{IF EXISTS LAST_TOPIC}
					<tr>
						<td>Last topic</td>
						<td>{LAST_TOPIC}</td>
					</tr>
					<tr>
						<td>Set by</td>
						<td>{LAST_TOPIC_SETTER}</td>
					</tr>
					{END IF}
					{IF EXISTS CAN_SET}
					<tr>
						<td>Keep topic</td>
						<td><input type="checkbox" name="keeptopic" value="on" {IF EXISTS KEEPTOPIC}checked{END IF}></td>
					</tr>
					<tr>
						<td>Peace</td>
						<td><input type="checkbox" name="peace" value="on" {IF EXISTS PEACE}checked{END IF}></td>
					</tr>
					<tr>
						<td>Private</td>
						<td><input type="checkbox" name="private" value="on" {IF EXISTS PRIVATE}checked{END IF}></td>
					</tr>
					<tr>
						<td>Restricted</td>
						<td><input type="checkbox" name="restricted" value="on" {IF EXISTS RESTRICTED}checked{END IF}></td>
					</tr>
					<tr>
						<td>Secure</td>
						<td><input type="checkbox" name="secure" value="on" {IF EXISTS SECURE}checked{END IF}></td>
					</tr>
					<tr>
						<td>Secure Ops</td>
						<td><input type="checkbox" name="secureops" value="on" {IF EXISTS SECUREOPS}checked{END IF}></td>
					</tr>
					<tr>
						<td>Topic Lock</td>
						<td><input type="checkbox" name="topiclock" value="on" {IF EXISTS TOPICLOCK}checked{END IF}></td>
					</tr>
					{IF EXISTS CAN_SET}
					<tr>
						<td></td>
						<td><button type="submit" class="btn btn-primary">Save</button></td>
					</tr>
					{END IF}
					{END IF}
				</table>

				{IF EXISTS CAN_SET}
				</form>
				{END IF}
			{END IF}
		</div>
{INCLUDE footer.html}
