{INCLUDE header.html}
		<div class="panel-heading">Nick access list</div>
		<div class="panel-body">
			{FOR M IN MESSAGES}
			<div class="alert alert-info">
				{M}<br>
			</div>
			{END FOR}

			{IF EXISTS ACCESS}
			<table id="tableNSAccess" class="table table-hover">
				<tbody>
					{FOR A IN ACCESS}
					<tr>
						<td class="col-lg-4">{A}</td>
						<td class="col-lg-8"><a href="/nickserv/access?mask={A}&del=1" class="btn btn-sm btn-danger">Delete</a></td>
					</tr>
					{END FOR}
				</tbody>
			</table>
			{ELSE}
			<em>Your access list is empty.</em>
			{END IF}

			<hr>

			<h4>Add an access entry</h4>
			<form class="form-horizontal" method="post" action="/nickserv/access">
				<div class="form-group">
					<label class="control-label col-lg-2" for="message">Hostmask:</label>
					<div class="col-lg-5">
						<input class="form-control" type="text" name="message" id="message" placeholder="Mask must be in the form user@host">
					</div>
				</div>
				<div class="form-group">
					<div class="col-lg-offset-2 col-lg-5">
						<button type="submit" class="btn btn-primary">Add</button>
					</div>
				</div>
			</form>
		</div>
{INCLUDE footer.html}
