{INCLUDE header.html}
{INCLUDE chanserv/chanlist.html}
		<div class="panel-heading">Channels you can drop</div>
		<div class="panel-body">
			{FOR M IN MESSAGES}
			<div class="alert alert-info">
				{M}<br>
			</div>
			{END FOR}

			{IF EXISTS CHANNEL_DROP}
				<div class="alert alert-danger" style="margin-top: 15px; margin-bottom: 0; padding-bottom: 0;">
					<h4>Drop Channel <strong>{CHANNEL_DROP}</strong>?</h4>
					<form class="form-horizontal" method="post" action="/chanserv/drop">
						<input type="hidden" value="{CHANNEL_DROP}" name="channel">
						<input type="hidden" value="yes" name="drop">
						<div class="form-group">
							<label class="control-label col-lg-4" for="confChan">Confirm channel name:</label>
							<div class="col-lg-4">
								<input class="form-control" type="text" name="confChan" id="confChan" placeholder="This cannot be undone!">
							</div>
						</div>
						<div class="form-group">
							<div class="col-lg-offset-4 col-lg-4">
								<button type="submit" class="btn btn-danger">Drop</button>
							</div>
						</div>
					</form>
				</div>
			{END IF}
		</div>
{INCLUDE footer.html}