{INCLUDE header.html}
{INCLUDE chanserv/chanlist.html}
		<div class="panel-heading">Akick List</div>
		<div class="panel-body">
			{FOR M IN MESSAGES}
			<div class="alert alert-info">
				{M}<br>
			</div>
			{END FOR}

			{IF EQ AKICK YES}
				{IF EXISTS MASKS}
				<table id="tableNSAccess" class="table table-hover">
					<thead>
						<tr>
							<th>Mask</th>
							<th>Reason</th>
							<th>Creator</th>
							<th></th>
						</tr>
					</thead>
					<tbody>
					{FOR MASK,REASON,CREATOR IN MASKS,REASONS,CREATORS}
						<tr>
							<td style="font-weight: bold;">{MASK}</td>
							<td>{REASON}</td>
							<td>{CREATOR}</td>
							<td><a href="/chanserv/akick?channel={ESCAPED_CHANNEL}&mask={MASK}&del=1" class="btn btn-sm btn-danger">Delete</a></td>
						</tr>
					{END FOR}
					</tbody>
				</table>
				{ELSE}
				<em>Akick list is empty.</em>
				{END IF}

				<hr>

				<h4>Add an akick entry</h4>
				<form class="form-horizontal" method="post" action="/chanserv/akick?channel={ESCAPED_CHANNEL}">
					<div class="form-group">
						<label class="control-label col-lg-2" for="mask">Mask:</label>
						<div class="col-lg-6">
							<input class="form-control" type="text" name="mask" id="mask" placeholder="Mask must be in the form nick!user@host">
						</div>
					</div>
					<div class="form-group">
						<label class="control-label col-lg-2" for="reason">Reason:</label>
						<div class="col-lg-6">
							<input class="form-control" type="text" name="reason" id="reason" placeholder="Reason for AKICK">
						</div>
					</div>
					<div class="form-group">
						<div class="col-lg-offset-2 col-lg-6">
							<button type="submit" class="btn btn-primary">Send</button>
						</div>
					</div>
				</form>
			{END IF}
		</div>
{INCLUDE footer.html}
