<!DOCTYPE html>
<!--[if IE 7]> <html lang="en" class="ie7"> <![endif]-->
<!--[if IE 8]> <html lang="en" class="ie8"> <![endif]-->
<!--[if IE 9]> <html lang="en" class="ie9"> <![endif]-->
<!--[if !IE]><!-->
<html lang="en">
<!--<![endif]-->
	<head>
		<link href="//netdna.bootstrapcdn.com/bootstrap/3.0.0/css/bootstrap.min.css" rel="stylesheet">
		<link href="/static/style.css" media="screen" rel="stylesheet" type="text/css" />

		<script src="//ajax.googleapis.com/ajax/libs/jquery/1.10.2/jquery.min.js"></script>
		<script src="//netdna.bootstrapcdn.com/bootstrap/3.0.0/js/bootstrap.min.js"></script>

		<!--[if lt IE 9]>
			<script src="//cdnjs.cloudflare.com/ajax/libs/html5shiv/3.6.2/html5shiv.js"></script>
			<script src="//cdnjs.cloudflare.com/ajax/libs/respond.js/1.2.0/respond.js"></script>
		<![endif]-->

		<meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
		<title>{TITLE}</title>
	</head>
	<body>
		<div id="frontPages" class="container">
			<div class="navbar navbar-default">
				<span class="navbar-brand" style="margin-top: -1px">Anope Web <small>Control Panel</small></span>
				<ul class="nav navbar-nav">
					{FOR CATEGORY_URL,CATEGORY_NAME IN CATEGORY_URLS,CATEGORY_NAMES}
					<li
						{IF EQ CATEGORY NickServ}{IF EQ CATEGORY_NAME NickServ}class="active"{END IF}{END IF}
						{IF EQ CATEGORY ChanServ}{IF EQ CATEGORY_NAME ChanServ}class="active"{END IF}{END IF}
						{IF EQ CATEGORY MemoServ}{IF EQ CATEGORY_NAME MemoServ}class="active"{END IF}{END IF}
						{IF EQ CATEGORY HostServ}{IF EQ CATEGORY_NAME HostServ}class="active"{END IF}{END IF}
						{IF EQ CATEGORY OperServ}{IF EQ CATEGORY_NAME OperServ}class="active"{END IF}{END IF}>
						{IF EQ CATEGORY_NAME OperServ}
							{IF EXISTS IS_OPER}
								<a href="{CATEGORY_URL}">{CATEGORY_NAME}</a>
							{END IF}
						{ELSE}
							<a href="{CATEGORY_URL}">{CATEGORY_NAME}</a>
						{END IF}
					</li>
					{END FOR}
				</ul>
				<span id="loggedIn" class="pull-right">
					<span>Logged in as {ACCOUNT}</span> <a href="/logout">(Logout)</a>
				</span>
			</div>

			<div class="row">
				<div class="col-lg-3">
					<div id="navPanel" class="sidebar-nav-fixed">
						<div class="well" style="padding: 0;">
							<ul class="nav nav-list" style="padding: 10px;">
								{FOR SUBCATEGORY_URL,SUBCATEGORY_GET,SUBCATEGORY_NAME IN SUBCATEGORY_URLS,SUBCATEGORY_GETS,SUBCATEGORY_NAMES}
								<li><a href="{SUBCATEGORY_URL}{SUBCATEGORY_GET}">{SUBCATEGORY_NAME}</a></li>
								{END FOR}
							</ul>
						</div>
					</div>
				</div>

				<div class="col-lg-9">
					<div class="panel panel-default">
