{INCLUDE header.html}
		<div class="panel-heading">Akill List</div>
		<div class="panel-body">
			{IF EXISTS NOACCESS}
			<h2>Access denied.</h2>
			{ELSE}

			{FOR M IN MESSAGES}
			<div class="alert alert-info">
				{M}<br>
			</div>
			{END FOR}

			<script>
				$(function () \{
					$('.table tbody tr td abbr').popover();
				\});
			</script>

			{IF EXISTS AKILLS}
			<em>{AKILLS}</em>
			{ELSE}
			<table id="tableNSAccess" class="table table-hover">
				<thead>
					<tr>
						<th>Number</th>
						<th>Hostmask <small>(hover: Reason)</small></th>
						<th>Expires <small>(hover: Set Date)</small></th>
						<th>Setter</th>
						<th></th>
					</tr>
				</thead>
				<tbody>
					{FOR N,H,S,T,E,R IN NUMBER,HOST,SETTER,TIME,EXPIRE,REASON}
					<tr>
						<td>{N}</td>
						<td><abbr data-container="body" data-trigger="hover" data-placement="top" data-content="{R}">{H}</abbr></td>
						<td><abbr data-container="body" data-trigger="hover" data-placement="top" data-content="{T}">{E}</abbr></td>
						<td>{S}</td>
						<td><a href="/operserv/akill?&number={N}&del=1" class="btn btn-sm btn-danger">Delete</a></td>
					</tr>
					{END FOR}
				</tbody>
			</table>
			{END IF}

			<hr>

			<h4>Add a new Akill</h4>
			<form class="form-horizontal" method="post" action="/operserv/akill">
				<div class="form-group">
					<label class="control-label col-lg-2" for="mask">HostMask:</label>
					<div class="col-lg-6">
						<input class="form-control" type="text" name="mask" id="mask" placeholder="Mask must be in the form nick!user@host">
					</div>
				</div>
				<div class="form-group">
					<label class="control-label col-lg-2" for="expiry">Expiry:</label>
					<div class="col-lg-6">
						<input class="form-control" type="text" name="expiry" id="expiry" placeholder="Expire time">
					</div>
				</div>
				<div class="form-group">
					<label class="control-label col-lg-2" for="reason">Message:</label>
					<div class="col-lg-6">
						<input class="form-control" type="text" name="reason" id="reason" placeholder="Reason for AKILL">
					</div>
				</div>
				<div class="form-group">
					<div class="col-lg-offset-2 col-lg-6">
						<button type="submit" class="btn btn-primary">Send</button>
					</div>
				</div>
			</form>
			{END IF}
		</div>
{INCLUDE footer.html}
