{INCLUDE header.html}
		<div class="panel-heading">Channel access list</div>
		<div class="panel-body">
			<table id="tableInfoNorm" class="table table-hover">
				<thead>
					<tr>
						<th>Number</th>
						<th>Channel</th>
						<th>Access</th>
					</tr>
				</thead>
				<tbody>
					{FOR N,C,A IN NUMBERS,CHANNELS,ACCESSES}
						<tr>
							<td>{N}</td>
							<td>{C}</td>
							<td>{A}</td>
						</tr>
					{END FOR}
				</tbody>
			</table>
		</div>
{INCLUDE footer.html}
