{INCLUDE header.html}
		<div class="panel-heading">vHost Information</div>
		<div class="panel-body">
			{FOR M IN MESSAGES}
			<div class="alert alert-info">
				{M}<br>
			</div>
			{END FOR}

			<table id="tableInfo" class="table table-hover">
				<tbody>
					<tr>
						<td>Your current vHost:</td>
						<td>
					{IF EXISTS VHOST}
						{VHOST}
					{ELSE}
						<em>None</em>
					{END IF}
						</td>
						<td></td>
					</tr>
					{IF EXISTS CAN_REQUEST}
					<tr>
						<td>
						{IF EXISTS VHOST}
							Request a new vHost
						{ELSE}
							Request a vHost
						{END IF}
						</td>
						<form method="post" action="/hostserv/request">
						<td><input class="form-control" name="req"></td>
						<td><button type="submit" class="btn btn-primary">Request</button></td>
						</form>
					</tr>
					{ELSE}
					<tr>
						<td colspan="2" class="text-center"><strong>vHost requests are disabled on this network.</strong></td>
					</tr>
					{END IF}
				</tbody>
			</table>
		</div>
{INCLUDE footer.html}
